@@ -30,6 +30,8 @@ pub enum Error {
     ///
     /// *Use HTTP/2 instead, or check if the [`ResponsePipe`] is HTTP/1*.
     PushOnHttp1,
+    /// Client closed connection before the response could be sent.
+    ClientRefusedResponse,
 }
 impl From<utils::parse::Error> for Error {
     #[inline]
@@ -40,6 +42,9 @@ impl From<utils::parse::Error> for Error {
 impl From<io::Error> for Error {
     #[inline]
     fn from(err: io::Error) -> Self {
+        if let io::ErrorKind::BrokenPipe = err.kind() {
+            return Self::ClientRefusedResponse;
+        }
         Self::Io(err)
     }
 }
@@ -47,6 +52,9 @@ impl From<io::Error> for Error {
 impl From<h2::Error> for Error {
     #[inline]
     fn from(err: h2::Error) -> Self {
+        if !err.is_io() && err.reason().is_none() {
+            return Self::ClientRefusedResponse;
+        }
         Self::H2(err)
     }
 }
@@ -66,6 +74,9 @@ impl From<Error> for io::Error {
                 io::ErrorKind::InvalidInput,
                 "can not push requests on http/1",
             ),
+            Error::ClientRefusedResponse => {
+                io::Error::new(io::ErrorKind::ConnectionReset, "client refused response")
+            }
         }
     }
 }
@@ -90,6 +101,28 @@ pub enum HttpConnection {
     Http2(Box<h2::server::Connection<Encryption, bytes::Bytes>>),
 }
 
+/// The data for [`Body::Bytes`].
+#[derive(Debug)]
+#[must_use]
+pub struct ByteBody {
+    content: Bytes,
+    read: usize,
+}
+impl ByteBody {
+    /// Get a reference to the bytes of this body.
+    pub fn inner(&self) -> &Bytes {
+        &self.content
+    }
+}
+impl From<Bytes> for ByteBody {
+    fn from(b: Bytes) -> Self {
+        Self {
+            content: b,
+            read: 0,
+        }
+    }
+}
+
 /// A body of a [`Request`].
 ///
 /// The inner variables are streams. To get the bytes, use [`Body::read_to_bytes()`] when needed.
@@ -97,10 +130,11 @@ pub enum HttpConnection {
 /// Also see [`FatRequest`].
 #[derive(Debug)]
 pub enum Body {
-    /// An empty body.
+    /// A body of [`Bytes`].
     ///
-    /// Can be used by HTTP/2 push to simulate a GET request.
-    Empty,
+    /// Can be used by HTTP/2 push to simulate a GET request,
+    /// or any other extensions which wants a Kvarn response.
+    Bytes(ByteBody),
     /// A buffered HTTP/1 body.
     ///
     /// While the HTTP/1 headers were read, it reads too much
@@ -164,6 +198,7 @@ impl HttpConnection {
     ///
     /// Also passes errors from [`h2::server::handshake`].
     pub async fn new(stream: Encryption, version: Version) -> Result<Self, Error> {
+        #[allow(clippy::match_same_arms)] // When http2 isn't enabled
         match version {
             Version::HTTP_09 | Version::HTTP_10 | Version::HTTP_11 => {
                 Ok(Self::Http1(Arc::new(Mutex::new(stream))))
@@ -176,7 +211,7 @@ impl HttpConnection {
                     .await;
                 match result {
                     Ok(connection) => Ok(HttpConnection::Http2(Box::new(connection))),
-                    Err(err) => Err(Error::H2(err)),
+                    Err(err) => Err(err.into()),
                 }
             }
             #[cfg(not(feature = "http2"))]
@@ -205,11 +240,11 @@ impl HttpConnection {
             }
             #[cfg(feature = "http2")]
             Self::Http2(connection) => match connection.accept().await {
-                Some(connection) => match connection {
+                Some(exchange) => match exchange {
                     Ok((request, response)) => {
                         Ok((request.map(Body::Http2), ResponsePipe::Http2(response)))
                     }
-                    Err(err) => Err(Error::H2(err)),
+                    Err(err) => Err(err.into()),
                 },
                 None => Err(utils::parse::Error::Done.into()),
             },
@@ -236,7 +271,14 @@ mod request {
         };
         let lock = stream.lock().await;
 
-        let (head, bytes) = read::request(lock, max_len, default_host, scheme).await?;
+        let (head, bytes) = read::request(
+            lock,
+            max_len,
+            default_host,
+            scheme,
+            std::time::Duration::from_secs(5),
+        )
+        .await?;
         let body = Body::Http1(response::Http1Body::new(
             stream,
             bytes,
@@ -255,13 +297,21 @@ mod request {
         #[inline]
         pub async fn read_to_bytes(&mut self) -> io::Result<Bytes> {
             match self {
-                Self::Empty => Ok(Bytes::new()),
+                Self::Bytes(bytes) => Ok(bytes.inner().clone()),
                 Self::Http1(h1) => h1.read_to_bytes().await,
                 #[cfg(feature = "http2")]
-                Self::Http2(h2) => futures::future::poll_fn(|cx| h2.poll_data(cx))
-                    .await
-                    .unwrap_or_else(|| Ok(Bytes::new()))
-                    .map_err(|err| io::Error::new(io::ErrorKind::InvalidData, err)),
+                Self::Http2(h2) => {
+                    let mut bytes = bytes::BytesMut::new();
+                    while let Some(result) = h2.data().await {
+                        let data = result
+                            .map_err(|err| io::Error::new(io::ErrorKind::InvalidData, err))?;
+
+                        bytes.extend_from_slice(&data);
+
+                        h2.flow_control().release_capacity(data.len()).unwrap();
+                    }
+                    Ok(bytes.freeze())
+                }
             }
         }
     }
@@ -292,7 +342,17 @@ mod request {
                     }
                     Poll::Ready(Ok(()))
                 }
-                Self::Empty => Poll::Ready(Ok(())),
+                Self::Bytes(byte_body) => {
+                    let rest = byte_body.inner().get(byte_body.read..).unwrap_or(&[]);
+                    if rest.is_empty() {
+                        return Poll::Ready(Ok(()));
+                    }
+                    let len = std::cmp::min(buf.remaining(), rest.len());
+                    buf.put_slice(&rest[..len]);
+                    byte_body.read += len;
+                    cx.waker().wake_by_ref();
+                    Poll::Pending
+                }
             }
         }
     }
@@ -311,6 +371,7 @@ mod response {
         offset: usize,
 
         content_length: usize,
+        // also update Debug implementation when adding fields
     }
     impl<R: AsyncRead + Unpin> Http1Body<R> {
         /// Creates a new body.
@@ -344,7 +405,7 @@ mod response {
             } else {
                 buffer.extend(&self.bytes);
                 if let Ok(result) = timeout(
-                    time::Duration::from_millis(250),
+                    Duration::from_millis(250),
                     async_bits::read_to_end_or_max(&mut buffer, &mut *self, len),
                 )
                 .await
@@ -366,12 +427,19 @@ mod response {
     }
     impl<R: AsyncRead + Unpin + Debug> Debug for Http1Body<R> {
         fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
-            f.debug_struct("Http1Body")
-                .field("reader", &self.reader)
-                .field("buffer", &"[internal buffer]".as_clean())
-                .field("offset", &self.offset)
-                .field("content_length", &self.content_length)
-                .finish()
+            let mut s = f.debug_struct(utils::ident_str!(
+                Http1Body,
+                R,
+                R: AsyncRead + Unpin + Debug
+            ));
+            utils::fmt_fields!(
+                s,
+                (self.reader),
+                (self.bytes, &"[internal buffer]".as_clean()),
+                (self.offset),
+                (self.content_length)
+            );
+            s.finish()
         }
     }
     impl<R: AsyncRead + Unpin> AsyncRead for Http1Body<R> {
@@ -441,17 +509,18 @@ mod response {
                         _ => {}
                     }
                     let mut writer = tokio::io::BufWriter::with_capacity(512, &mut *writer);
-                    async_bits::write::response(&response, b"", &mut writer)
-                        .await
-                        .map_err(Error::Io)?;
-                    writer.flush().await.map_err(Error::Io)?;
+                    async_bits::write::response(&response, b"", &mut writer).await?;
+                    writer.flush().await?;
                     writer.into_inner();
 
                     Ok(ResponseBodyPipe::Http1(Arc::clone(s)))
                 }
                 #[cfg(feature = "http2")]
                 Self::Http2(s) => match s.send_response(response, end_of_stream) {
-                    Err(err) => Err(Error::H2(err)),
+                    Err(ref err) if err.get_io().is_none() && err.reason().is_none() => {
+                        Err(Error::ClientRefusedResponse)
+                    }
+                    Err(err) => Err(err.into()),
                     Ok(pipe) => Ok(ResponseBodyPipe::Http2(pipe)),
                 },
             }
@@ -463,6 +532,7 @@ mod response {
         /// If you try to push if `self` is [`ResponsePipe::Http1`], an [`Error::PushOnHttp1`] is returned.
         /// Returns errors from [`h2::server::SendResponse::push_request()`].
         #[inline]
+        #[allow(clippy::needless_pass_by_value)]
         pub fn push_request(
             &mut self,
             #[allow(unused_variables)] request: Request<()>,
@@ -472,7 +542,7 @@ mod response {
                 #[cfg(feature = "http2")]
                 Self::Http2(h2) => match h2.push_request(request) {
                     Ok(pipe) => Ok(PushedResponsePipe::Http2(pipe)),
-                    Err(err) => Err(Error::H2(err)),
+                    Err(err) => Err(err.into()),
                 },
             }
         }
@@ -501,6 +571,7 @@ mod response {
         /// Errors are passed from the HTTP libraries, for now only [`mod@h2`].
         /// See [`h2::server::SendPushedResponse::send_response()`] for more information.
         #[inline]
+        #[allow(clippy::needless_pass_by_value)]
         pub fn send_response(
             &mut self,
             response: Response<()>,
@@ -513,7 +584,7 @@ mod response {
                     *response.version_mut() = Version::HTTP_2;
 
                     match s.send_response(response, end_of_stream) {
-                        Err(err) => Err(Error::H2(err)),
+                        Err(err) => Err(err.into()),
                         Ok(pipe) => Ok(ResponseBodyPipe::Http2(pipe)),
                     }
                 }
@@ -574,7 +645,7 @@ mod response {
         #[inline]
         pub async fn close(&mut self) -> Result<(), Error> {
             match self {
-                Self::Http1(h1) => h1.lock().await.flush().await.map_err(Error::from),
+                Self::Http1(h1) => h1.lock().await.flush().await.map_err(Into::into),
                 #[cfg(feature = "http2")]
                 Self::Http2(h2) => h2.send_data(Bytes::new(), true).map_err(Error::from),
             }
