@@ -13,10 +13,13 @@ use std::{
     hash::{Hash, Hasher},
 };
 
+/// The HTTP date time format in the [`time`] format.
+pub static HTTP_DATE: &[time::format_description::FormatItem] = time::macros::format_description!("[weekday repr:short case_sensitive:true], [day padding:zero] [month repr:short case_sensitive:true] [year padding:zero repr:full base:calendar sign:automatic] [hour repr:24 padding:zero]:[minute padding:zero]:[second padding:zero] GMT");
+
 /// A [`Cache`] inside a [`Mutex`] with appropriate type parameters for a file cache.
-pub type FileCache = Mutex<Cache<PathBuf, Bytes>>;
+pub type FileCache = RwLock<Cache<PathBuf, Bytes>>;
 /// A [`Cache`] inside a [`Mutex`] with appropriate type parameters for a response cache.
-pub type ResponseCache = Mutex<Cache<UriKey, VariedResponse>>;
+pub type ResponseCache = RwLock<Cache<UriKey, VariedResponse>>;
 
 /// A path an optional query used in [`UriKey`]
 ///
@@ -165,6 +168,87 @@ pub fn do_compress(mime: &Mime, check_utf8: impl Fn() -> bool) -> bool {
                 || check_utf8()))
 }
 
+/// The preferred compression algorithm.
+///
+/// The default is chosen according to the [cargo features](https://kvarn.org/cargo-features.) in
+/// the following order:
+/// - Brotli
+/// - Gzip
+/// - None
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+#[non_exhaustive]
+pub enum PreferredCompression {
+    /// Prefer the brotli algorithm.
+    ///
+    /// This is the default and is the best.
+    #[cfg(feature = "br")]
+    Brotli,
+    /// Prefer the gzip algorithm.
+    ///
+    /// Uses a bit less memory than [`Self::Brotli`] at the expense of compression.
+    #[cfg(feature = "gzip")]
+    Gzip,
+    /// Prefer no compression. This is the default if no compression features are enabled.
+    None,
+}
+impl PreferredCompression {
+    /// Return the name used in the `content-encoding` header.
+    #[must_use]
+    pub fn as_str(&self) -> &'static str {
+        match self {
+            #[cfg(feature = "br")]
+            Self::Brotli => "br",
+            #[cfg(feature = "gzip")]
+            Self::Gzip => "gzip",
+            Self::None => "identity",
+        }
+    }
+}
+impl Default for PreferredCompression {
+    fn default() -> Self {
+        #[cfg(feature = "br")]
+        {
+            PreferredCompression::Brotli
+        }
+        #[cfg(all(not(feature = "br"), feature = "gzip"))]
+        {
+            PreferredCompression::Gzip
+        }
+        #[cfg(all(not(feature = "br"), not(feature = "gzip")))]
+        {
+            PreferredCompression::None
+        }
+    }
+}
+/// Some options for how to compress the response.
+#[derive(Debug, Clone)]
+pub struct CompressionOptions {
+    /// The preferred compression algorithm.
+    pub preferred: PreferredCompression,
+    /// The level of brotli compression.
+    ///
+    /// See [some benchmarks](https://quixdb.github.io/squash-benchmark/#results) for more context.
+    #[cfg(feature = "br")]
+    pub brotli_level: u32,
+    /// The level of gzip compression.
+    ///
+    /// See [some benchmarks](https://quixdb.github.io/squash-benchmark/#results) for more context.
+    #[cfg(feature = "gzip")]
+    pub gzip_level: u32,
+}
+#[allow(clippy::derivable_impls)] // if no features are enabled, we get a warning, since the only field is using it's default
+impl Default for CompressionOptions {
+    fn default() -> Self {
+        Self {
+            preferred: PreferredCompression::default(),
+            #[cfg(feature = "br")]
+            brotli_level: 3,
+            #[cfg(feature = "gzip")]
+            gzip_level: 1,
+        }
+    }
+}
+
 /// A response with a lazily compressed body.
 ///
 /// The compressed body is cached.
@@ -174,7 +258,9 @@ pub fn do_compress(mime: &Mime, check_utf8: impl Fn() -> bool) -> bool {
 #[must_use]
 pub struct CompressedResponse {
     identity: Response<Bytes>,
+    #[cfg(feature = "gzip")]
     gzip: Option<Bytes>,
+    #[cfg(feature = "br")]
     br: Option<Bytes>,
 
     compress: CompressPreference,
@@ -193,7 +279,9 @@ impl CompressedResponse {
         Self::check_content_type(&mut identity, extension);
         Self {
             identity,
+            #[cfg(feature = "gzip")]
             gzip: None,
+            #[cfg(feature = "br")]
             br: None,
 
             compress,
@@ -217,6 +305,7 @@ impl CompressedResponse {
     pub fn clone_preferred<T>(
         &self,
         request: &Request<T>,
+        options: &CompressionOptions,
     ) -> Result<Response<Bytes>, &'static str> {
         let values = match request
             .headers()
@@ -232,20 +321,6 @@ impl CompressedResponse {
             .iter()
             .any(|v| v.value == "identity" && v.quality == 0.0);
 
-        #[cfg(all(feature = "gzip", feature = "br"))]
-        let prefer_br = values
-            .iter()
-            .find(|v| v.value == "gzip")
-            .map_or(0.0, |v| v.quality)
-            <= 0.5
-            && values.iter().find_map(|v| {
-                if v.value == "br" {
-                    Some(v.quality)
-                } else {
-                    None
-                }
-            }) == Some(1.0);
-
         let only_identity = values.len() == 1
             && values[0]
                 == utils::ValueQualitySet {
@@ -274,37 +349,35 @@ impl CompressedResponse {
         let (bytes, compression) = match &mime {
             Some(mime) => {
                 if do_compress(mime, || str::from_utf8(self.get_identity().body()).is_ok()) {
-                    // You are wrong. We have `true if ...`.
-                    #[allow(clippy::match_bool)]
                     match self.compress {
                         CompressPreference::None => (self.get_identity().body(), "identity"),
                         CompressPreference::Full => {
-                            #[cfg(all(feature = "gzip", feature = "br"))]
-                            match (self.br.is_some() && self.gzip.is_none()) || prefer_br {
-                                true if contains("br") => (self.get_br(), "br"),
-                                true if contains("gzip") => (self.get_gzip(), "gzip"),
-                                false if contains("gzip") => (self.get_gzip(), "gzip"),
-                                false if contains("br") => (self.get_br(), "br"),
-                                _ => (self.get_identity().body(), "identity"),
-                            }
-                            #[cfg(all(feature = "gzip", not(feature = "br")))]
-                            {
-                                match contains("gzip") {
-                                    true => (self.get_gzip(), "gzip"),
-                                    false => (self.get_identity().body(), "identity"),
+                            #[cfg(feature = "br")]
+                            let contains_br = contains("br");
+                            #[cfg(feature = "gzip")]
+                            let contains_gzip = contains("gzip");
+
+                            #[allow(unused_mut)]
+                            let mut preferred = match options.preferred.as_str() {
+                                #[cfg(feature = "br")]
+                                "br" if contains_br => {
+                                    Some((self.get_br(options.brotli_level), "br"))
                                 }
-                            }
-                            #[cfg(all(feature = "br", not(feature = "gzip")))]
-                            {
-                                match contains("br") {
-                                    true => (self.get_br(), "br"),
-                                    false => (self.get_identity().body(), "identity"),
+                                #[cfg(feature = "gzip")]
+                                "gzip" if contains_gzip => {
+                                    Some((self.get_gzip(options.gzip_level), "gzip"))
                                 }
+                                _ => None,
+                            };
+                            #[cfg(feature = "br")]
+                            if preferred.is_none() && contains_br {
+                                preferred = Some((self.get_br(options.brotli_level), "br"));
                             }
-                            #[cfg(not(any(feature = "gzip", feature = "br")))]
-                            {
-                                (self.get_identity().body(), "identity")
+                            #[cfg(feature = "gzip")]
+                            if preferred.is_none() && contains_gzip {
+                                preferred = Some((self.get_gzip(options.gzip_level), "gzip"));
                             }
+                            preferred.unwrap_or_else(|| (self.get_identity().body(), "identity"))
                         }
                     }
                 } else {
@@ -345,12 +418,17 @@ impl CompressedResponse {
     }
     fn check_content_type(response: &mut Response<Bytes>, extension: &str) {
         fn add_utf_8(headers: &mut HeaderMap, mime: &Mime) {
+            let charset = if mime.get_param(mime::CHARSET) == Some(mime::UTF_8) {
+                ""
+            } else {
+                "; charset=utf-8"
+            };
             // We know the added bytes are safe for a http::HeaderValue
             // unwrap is ok.
-            let content_type = HeaderValue::from_maybe_shared(Bytes::copy_from_slice(
-                format!("{}; charset=utf-8", mime).as_bytes(),
-            ))
-            .unwrap();
+            let mut header = mime.to_string();
+            header.push_str(charset);
+            let content_type =
+                HeaderValue::from_maybe_shared::<Bytes>(header.into_bytes().into()).unwrap();
             utils::replace_header(headers, "content-type", content_type);
         }
         let utf_8 = response.body().len() < 16 * 1024 && str::from_utf8(response.body()).is_ok();
@@ -388,9 +466,9 @@ impl CompressedResponse {
                     add_utf_8(response.headers_mut(), &mime_type);
                 } else {
                     // Mime will only contains valid bytes.
-                    let content_type = HeaderValue::from_maybe_shared(Bytes::copy_from_slice(
-                        mime_type.to_string().as_bytes(),
-                    ))
+                    let content_type = HeaderValue::from_maybe_shared::<Bytes>(
+                        mime_type.to_string().into_bytes().into(),
+                    )
                     .unwrap();
                     response.headers_mut().insert("content-type", content_type);
                 }
@@ -426,13 +504,14 @@ impl CompressedResponse {
     /// You should use [`Self::clone_preferred`] to get the preferred compression instead,
     /// as it is available with any set of features
     #[cfg(feature = "gzip")]
-    pub fn get_gzip(&self) -> &Bytes {
+    pub fn get_gzip(&self, level: u32) -> &Bytes {
         if self.gzip.is_none() {
             let bytes = self.identity.body().as_ref();
 
-            let mut buffer = utils::WriteableBytes::with_capacity(bytes.len() / 2 + 64);
+            let mut buffer = utils::WriteableBytes::with_capacity(bytes.len() / 3 + 64);
 
-            let mut c = flate2::write::GzEncoder::new(&mut buffer, flate2::Compression::fast());
+            // 1-9, 1 is fast, 9 is slow. 4 is equal to brotli's 3
+            let mut c = flate2::write::GzEncoder::new(&mut buffer, flate2::Compression::new(level));
             c.write_all(bytes).expect("Failed to compress using gzip!");
             c.finish().expect("Failed to compress using gzip!");
 
@@ -453,13 +532,14 @@ impl CompressedResponse {
     /// You should use [`Self::clone_preferred`] to get the preferred compression instead,
     /// as it is available with any set of features
     #[cfg(feature = "br")]
-    pub fn get_br(&self) -> &Bytes {
+    pub fn get_br(&self, level: u32) -> &Bytes {
         if self.br.is_none() {
             let bytes = self.identity.body().as_ref();
 
-            let mut buffer = utils::WriteableBytes::with_capacity(bytes.len() / 2 + 64);
+            let mut buffer = utils::WriteableBytes::with_capacity(bytes.len() / 3 + 64);
 
-            let mut c = brotli::CompressorWriter::new(&mut buffer, 4096, 8, 21);
+            // 1-10, 1 is fast, 10 is really slow
+            let mut c = brotli::CompressorWriter::new(&mut buffer, 4096, level, 21);
             c.write_all(bytes)
                 .expect("Failed to compress using Brotli!");
             c.flush().expect("Failed to compress using Brotli!");
@@ -499,8 +579,8 @@ pub enum CompressPreference {
 /// assert_eq!(Err(CachePreferenceError::Invalid), "FULL".parse::<ClientCachePreference>());
 /// assert_eq!(Ok(ServerCachePreference::QueryMatters), "query-matters".parse());
 /// use std::convert::TryInto;
-/// assert_eq!(Ok(ClientCachePreference::MaxAge(time::Duration::from_secs(42))), "42s".parse());
-/// assert_eq!(Ok(ServerCachePreference::MaxAge(time::Duration::from_secs(3600))), "3600s".parse());
+/// assert_eq!(Ok(ClientCachePreference::MaxAge(Duration::from_secs(42))), "42s".parse());
+/// assert_eq!(Ok(ServerCachePreference::MaxAge(Duration::from_secs(3600))), "3600s".parse());
 /// assert_eq!(Err(CachePreferenceError::ZeroDuration), "0s".parse::<ClientCachePreference>());
 /// assert_eq!(Err(CachePreferenceError::ZeroDuration), "0s".parse::<ServerCachePreference>());
 /// ```
@@ -531,24 +611,21 @@ pub enum ServerCachePreference {
     Full,
     /// Sets a max age for the content.
     /// Query will be discarded in cache, same as [`Self::Full`].
-    MaxAge(time::Duration),
+    MaxAge(Duration),
 }
 impl ServerCachePreference {
-    /// If `_response` should be cached.
-    ///
-    /// Checks [`Self`] for preference and [`Response::status()`].
-    /// Will not cache status codes between 400..=403 and 405..=499
+    /// If a response with `cache_action` (from [`host::CacheAction`]) should be cached.
     #[inline]
     #[must_use]
     #[allow(clippy::unused_self)]
-    pub fn cache(self, _response: &Response<Bytes>, _method: &Method) -> bool {
+    pub fn cache(self, cache_action: host::CacheAction, method: &Method) -> bool {
         let of_self = match self {
             Self::None => false,
             Self::QueryMatters | Self::Full | Self::MaxAge(_) => true,
         };
-        #[allow(clippy::unnested_or_patterns)]
-        let of_response = !matches!(_response.status().as_u16(), 400..=403 | 405..=499)
-            && matches!(_method, &Method::GET | &Method::HEAD);
+        #[allow(clippy::unnested_or_patterns)] // matches! macro
+        let of_response =
+            cache_action.into_cache() && matches!(method, &Method::GET | &Method::HEAD);
         of_self && of_response
     }
     /// If query matters in cache.
@@ -575,11 +652,11 @@ impl str::FromStr for ServerCachePreference {
             "" => return Err(CachePreferenceError::Empty),
             _ => {
                 if let Some(integer) = s.strip_suffix('s') {
-                    if let Ok(integer) = integer.parse() {
+                    if let Ok(integer) = integer.parse::<u64>() {
                         if integer == 0 {
                             return Err(CachePreferenceError::ZeroDuration);
                         }
-                        return Ok(Self::MaxAge(time::Duration::from_secs(integer)));
+                        return Ok(Self::MaxAge(integer.std_seconds()));
                     }
                 }
                 return Err(CachePreferenceError::Invalid);
@@ -602,7 +679,7 @@ pub enum ClientCachePreference {
     ///
     /// Note that this must be in seconds when sending a header.
     /// This will be rounded up.
-    MaxAge(time::Duration),
+    MaxAge(Duration),
 }
 impl ClientCachePreference {
     /// Gets the [`HeaderValue`] representation of the preference.
@@ -611,17 +688,19 @@ impl ClientCachePreference {
     pub fn as_header(self) -> HeaderValue {
         match self {
             Self::None => HeaderValue::from_static("no-store"),
-            Self::Changing => HeaderValue::from_static("max-age=120"),
+            Self::Changing => HeaderValue::from_static("max-age=2"),
             Self::Full => HeaderValue::from_static("public, max-age=604800, immutable"),
             Self::MaxAge(duration) => {
                 let bytes = build_bytes!(
                     b"public, max-age=",
+                    // if > second integer, add 1 second (ceil the duration).
+                    // i64::from(bool) returns 1 if true.
                     (duration.as_secs() + u64::from(duration.subsec_nanos() > 0))
                         .to_string()
                         .as_bytes(),
                     b", immutable"
                 );
-                // We know the bytes are safe.
+                // We know these bytes are safe.
                 HeaderValue::from_maybe_shared(bytes).unwrap()
             }
         }
@@ -631,11 +710,11 @@ impl str::FromStr for ClientCachePreference {
     type Err = CachePreferenceError;
     fn from_str(s: &str) -> Result<Self, Self::Err> {
         if let Some(integer) = s.strip_suffix('s') {
-            if let Ok(integer) = integer.parse() {
+            if let Ok(integer) = integer.parse::<u64>() {
                 if integer == 0 {
                     return Err(CachePreferenceError::ZeroDuration);
                 }
-                return Ok(Self::MaxAge(time::Duration::from_secs(integer)));
+                return Ok(Self::MaxAge(integer.std_seconds()));
             }
         }
         Ok(match s {
@@ -694,10 +773,13 @@ impl<V> CacheOut<V> {
 /// `T` represents the cached data.
 ///
 /// The other information is for lifetimes of the cache.
-/// The [`DateTime`] is when the item was added and
+/// The [`OffsetDateTime`] is when the item was added and
 /// the [`Duration`] how long the item can be kept.
 /// A `Duration` value of `None` means the item will never expire.
-pub type CacheItem<T> = (T, (DateTime<Utc>, Option<Duration>));
+///
+/// Keep in mind that `Duration` is the std variant, while `Duration` is the time crate's
+/// variant, which supports negative durations.
+pub type CacheItem<T> = (T, (OffsetDateTime, Option<Duration>));
 
 /// A general cache with size and item count limits.
 ///
@@ -751,41 +833,72 @@ impl<K: Eq + Hash, V, H> Cache<K, V, H> {
     ///
     /// See [`HashMap::get`] for more info.
     #[inline]
-    pub fn get<Q: ?Sized + Hash + Eq>(&mut self, key: &Q) -> CacheOut<&mut V>
+    pub fn get<Q: ?Sized + Hash + Eq>(&self, key: &Q) -> CacheOut<&V>
     where
         K: Borrow<Q>,
     {
-        self.get_with_lifetime(key).map(|v| &mut v.0)
+        self.get_with_lifetime(key).map(|v| &v.0)
     }
     /// Gets the [`CacheItem`] at `key` from the cache.
     /// Consider using [`Self::get`] for most operations.
     ///
     /// This includes all lifetime information about the item in the cache.
     /// See [`CacheItem`] for more info about this.
-    pub fn get_with_lifetime<Q: ?Sized + Hash + Eq>(
+    pub fn get_with_lifetime<Q: ?Sized + Hash + Eq>(&self, key: &Q) -> CacheOut<&CacheItem<V>>
+    where
+        K: Borrow<Q>,
+    {
+        // `TODO`: maybe set tokio timers to remove items instead?
+        match self.map.get(key) {
+            Some(value_and_lifetime)
+                if value_and_lifetime.1 .1.map_or(true, |lifetime| {
+                    OffsetDateTime::now_utc() - value_and_lifetime.1 .0 <= lifetime
+                }) =>
+            {
+                CacheOut::Present(value_and_lifetime)
+            }
+            Some(_) | None => CacheOut::None,
+        }
+    }
+    /// Get a mutable reference to the value at `key` from the cache.
+    ///
+    /// See [`HashMap::get`] for more info.
+    #[inline]
+    pub fn get_mut<Q: ?Sized + Hash + Eq>(&mut self, key: &Q) -> CacheOut<&V>
+    where
+        K: Borrow<Q>,
+    {
+        self.get_mut_with_lifetime(key).map(|v| &v.0)
+    }
+    /// Gets a mutable reference to the [`CacheItem`] at `key` from the cache.
+    /// Consider using [`Self::get`] for most operations.
+    ///
+    /// This includes all lifetime information about the item in the cache.
+    /// See [`CacheItem`] for more info about this.
+    pub fn get_mut_with_lifetime<Q: ?Sized + Hash + Eq>(
         &mut self,
         key: &Q,
     ) -> CacheOut<&mut CacheItem<V>>
     where
         K: Borrow<Q>,
     {
-        // Here for borrowing issues after `self.map.get_mut`.
-        // See the SAFETY note bellow.
-        let ptr: *const _ = self;
+        // // Here for borrowing issues after `self.map.get_mut`.
+        // // See the SAFETY note bellow.
+        // let ptr: *const _ = self;
         // maybe set tokio timers to remove items instead?
         match self.map.get_mut(key) {
             Some(value_and_lifetime)
                 if value_and_lifetime.1 .1.map_or(true, |lifetime| {
-                    Utc::now() - value_and_lifetime.1 .0 <= lifetime
+                    OffsetDateTime::now_utc() - value_and_lifetime.1 .0 <= lifetime
                 }) =>
             {
                 CacheOut::Present(value_and_lifetime)
             }
             Some(_) => {
-                // SAFETY: No other have a reference to self; the other branches are just that,
-                // other branches, and their references are returned, so this code isn't ran.
-                #[allow(clippy::cast_ref_to_mut)]
-                unsafe { &mut *(ptr as *mut Cache<K, V>) }.remove(key);
+                // // SAFETY: No other have a reference to self; the other branches are just that,
+                // // other branches, and their references are returned, so this code isn't ran.
+                // #[allow(clippy::cast_ref_to_mut)]
+                // unsafe { &mut *(ptr as *mut Cache<K, V>) }.remove(key);
                 CacheOut::None
             }
             None => CacheOut::None,
@@ -835,7 +948,10 @@ impl<K: Eq + Hash, V, H> Cache<K, V, H> {
         if self.map.len() >= self.max_items {
             self.discard_one();
         }
-        match self.map.insert(key, (value, (Utc::now(), lifetime))) {
+        match self
+            .map
+            .insert(key, (value, (OffsetDateTime::now_utc(), lifetime)))
+        {
             Some((v, _expiry)) => CacheOut::Present(v),
             None => CacheOut::None,
         }
@@ -881,7 +997,7 @@ impl<K: Eq + Hash, H: Hasher> Cache<K, VariedResponse, H> {
                 .ok()
                 .as_ref()
                 .and_then(parse::CacheControl::as_freshness)
-                .map(|s| Duration::seconds(i64::from(s)));
+                .map(|s| u64::from(s).std_seconds());
 
         let identity = response.first().0.get_identity().body();
         let identity_fragment = &identity[identity.len().saturating_sub(512)..];
