@@ -19,7 +19,7 @@ use tokio_tls::{MidHandshake, TlsState, TlsStream};
 pub enum Encryption {
     /// A TLS encrypted TCP stream.
     #[cfg(feature = "https")]
-    TcpTls(TlsStream<TcpStream>),
+    TcpTls(Box<TlsStream<TcpStream>>),
     /// A unencrypted TCP stream for use with
     /// non-secure HTTP.
     Tcp(TcpStream),
@@ -49,7 +49,7 @@ impl Encryption {
                 let connect = acceptor.await.map_err(|(err, _)| err)?;
                 info!("Successful handshake");
 
-                Ok(Self::TcpTls(connect))
+                Ok(Self::TcpTls(Box::new(connect)))
             }
         }
     }
@@ -364,7 +364,7 @@ mod tokio_tls {
                         &mut self,
                         f: impl FnOnce(Pin<&mut T>, &mut Context<'_>) -> Poll<io::Result<U>>,
                     ) -> io::Result<U> {
-                        match f(Pin::new(&mut self.io), self.cx) {
+                        match f(Pin::new(self.io), self.cx) {
                             Poll::Ready(result) => result,
                             Poll::Pending => Err(io::ErrorKind::WouldBlock.into()),
                         }
@@ -383,7 +383,7 @@ mod tokio_tls {
                     }
 
                     fn flush(&mut self) -> io::Result<()> {
-                        self.poll_with(|io, cx| io.poll_flush(cx))
+                        self.poll_with(AsyncWrite::poll_flush)
                     }
                 }
 
