@@ -41,14 +41,18 @@ use rustls::{
 pub struct Host {
     /// The name of the host, will be used in matching the requests [SNI hostname](rustls::server::ClientHello::server_name())
     /// and `host` header to get the requested host to handle the request.
-    pub name: &'static str,
+    pub name: String,
+    /// The alternative names this host is recognized by.
+    /// This should probably be empty unless your certificate also covers these names.
+    pub alternative_names: Vec<String>,
     /// The certificate of this host, if any.
     #[cfg(feature = "https")]
     pub certificate: Option<Arc<sign::CertifiedKey>>,
     /// Base path of all data for this host.
     ///
     /// If you enabled the `fs` feature (enabled by default),
-    /// the public files are in the directory `<path>/public`.
+    /// the public files are in the directory `<path>/public`
+    /// (`public` by default; see [`Options::public_data_dir`]).
     ///
     /// Also, all extensions should use this to access data on disk.
     pub path: PathBuf,
@@ -75,10 +79,13 @@ pub struct Host {
 
     /// Other settings.
     pub options: Options,
+    /// Preferences and options for compression.
+    pub compression_options: comprash::CompressionOptions,
+    // also add to debug implementation when inserting new field
 }
 impl Host {
     /// Creates a new [`Host`].
-    /// Will read certificates in the specified locations
+    /// Will read PEM encoded certificates in the specified locations
     /// and return an non-secure host if parsing fails.
     ///
     /// To achieve greater security, use [`Host::with_http_to_https_redirect`] and call [`Host::with_hsts`].
@@ -91,7 +98,7 @@ impl Host {
     /// Will return any error from [`get_certified_key()`] with a [`Host`] containing no certificates.
     #[cfg(feature = "https")]
     pub fn try_read_fs(
-        host_name: &'static str,
+        host_name: impl AsRef<str>,
         cert_path: impl AsRef<Path>,
         private_key_path: impl AsRef<Path>,
         path: impl AsRef<Path>,
@@ -104,6 +111,34 @@ impl Host {
             Err(err) => Err((err, Self::unsecure(host_name, path, extensions, options))),
         }
     }
+    /// Same as [`Self::try_read_fs`], but extracts the host name from the certificate.
+    /// This also doesn't fall back to [`Self::unsecure`], as we don't know the name if the
+    /// certificate couldn't be parsed.
+    ///
+    /// # Panics
+    ///
+    /// Panics if the parsed certificate `.is_empty()` or if the first certificate in the parsed chain is invalid.
+    ///
+    /// # Errors
+    ///
+    /// Will return any error from [`get_certified_key()`].
+    #[cfg(all(feature = "https", feature = "auto-hostname"))]
+    pub fn read_fs_name_from_cert(
+        cert_path: impl AsRef<Path>,
+        private_key_path: impl AsRef<Path>,
+        path: impl AsRef<Path>,
+        extensions: Extensions,
+        options: Options,
+    ) -> Result<Self, CertificateError> {
+        let cert = get_certified_key(cert_path, private_key_path);
+        match cert {
+            Ok((cert, pk)) => Ok(Self::new_name_from_cert(
+                cert, pk, path, extensions, options,
+            )),
+            Err(err) => Err(err),
+        }
+    }
+
     /// Creates a new [`Host`] from the [`rustls`]
     /// `cert` and `pk`. When they are in files, consider [`Self::new`]
     /// which reads from files.
@@ -112,25 +147,26 @@ impl Host {
     ///
     /// # Examples
     ///
-    /// ```nocompile
+    /// ```ignore
+    /// # use kvarn::prelude::*;
     /// let certificate =
     ///     rcgen::generate_simple_self_signed(vec!["localhost".to_string()]).unwrap();
     /// let cert = vec![rustls::Certificate(certificate.serialize_der().unwrap())];
     /// let pk = rustls::PrivateKey(certificate.serialize_private_key_der());
     /// let pk = Arc::new(rustls::sign::any_supported_type(&pk).unwrap());
     ///
-    /// Host::from_cert_and_pk(
+    /// Host::new(
     ///     "localhost",
     ///     cert,
     ///     pk,
-    ///     PathBuf::from("tests"),
-    ///     extensions,
+    ///     "tests",
+    ///     Extensions::default(),
     ///     host::Options::default(),
-    /// )
+    /// );
     /// ```
     #[cfg(feature = "https")]
     pub fn new(
-        name: &'static str,
+        name: impl AsRef<str>,
         cert: Vec<rustls::Certificate>,
         pk: Arc<dyn sign::SigningKey>,
         path: impl AsRef<Path>,
@@ -140,38 +176,76 @@ impl Host {
         let cert = sign::CertifiedKey::new(cert, pk);
 
         Self {
-            name,
+            name: name.as_ref().to_owned(),
+            alternative_names: Vec::new(),
             certificate: Some(Arc::new(cert)),
             path: path.as_ref().to_path_buf(),
             extensions,
-            file_cache: Some(Mutex::new(Cache::default())),
-            response_cache: Some(Mutex::new(Cache::default())),
+            file_cache: Some(RwLock::new(Cache::default())),
+            response_cache: Some(RwLock::new(Cache::default())),
             options,
             limiter: LimitManager::default(),
             vary: Vary::default(),
+
+            compression_options: comprash::CompressionOptions::default(),
         }
     }
+    /// Same as [`Self::new`], but extracts the host name from the certificate.
+    ///
+    /// # Panics
+    ///
+    /// Panics if `cert.is_empty()` or if the first certificate in `cert` is invalid.
+    #[cfg(all(feature = "https", feature = "auto-hostname"))]
+    pub fn new_name_from_cert(
+        cert: Vec<rustls::Certificate>,
+        pk: Arc<dyn sign::SigningKey>,
+        path: impl AsRef<Path>,
+        extensions: Extensions,
+        options: Options,
+    ) -> Self {
+        use x509_parser::traits::FromDer;
+        let tbs = x509_parser::certificate::X509Certificate::from_der(&cert[0].0)
+            .expect("certificate invalid, failed to get host name")
+            .1
+            .tbs_certificate;
+        let name = tbs.subject().to_string();
+        let mut alt_names = Vec::new();
+        let alt_name = tbs
+            .subject_alternative_name()
+            .expect("alternative name extension of certificate invalid");
+        if let Some(alt_name) = alt_name {
+            for name in &alt_name.value.general_names {
+                alt_names.push(name.to_string());
+            }
+        }
+        let mut me = Self::new(name, cert, pk, path, extensions, options);
+        me.alternative_names = alt_names;
+        println!("Host: {}, alt: {:?}", me.name, me.alternative_names);
+        me
+    }
     /// Creates a new [`Host`] without a certificate.
     ///
     /// This host will only support non-encrypted HTTP/1 connections.
-    /// Consider enabling the `https` flag and use a self-signed certificate or one from [Let's Encrypt](https://letsencrypt.org/).
+    /// Consider enabling the `https` feature and use a self-signed certificate or one from [Let's Encrypt](https://letsencrypt.org/).
     pub fn unsecure(
-        host_name: &'static str,
+        host_name: impl AsRef<str>,
         path: impl AsRef<Path>,
         extensions: Extensions,
         options: Options,
     ) -> Self {
         Self {
-            name: host_name,
+            name: host_name.as_ref().to_owned(),
+            alternative_names: Vec::new(),
             #[cfg(feature = "https")]
             certificate: None,
             path: path.as_ref().to_path_buf(),
             extensions,
-            file_cache: Some(Mutex::new(Cache::default())),
-            response_cache: Some(Mutex::new(Cache::default())),
+            file_cache: Some(RwLock::new(Cache::default())),
+            response_cache: Some(RwLock::new(Cache::default())),
             options,
             limiter: LimitManager::default(),
             vary: Vary::default(),
+            compression_options: comprash::CompressionOptions::default(),
         }
     }
 
@@ -234,10 +308,14 @@ impl Host {
     /// Also see [hstspreload.org](https://hstspreload.org/)
     #[cfg(feature = "https")]
     pub fn with_hsts(&mut self) -> &mut Self {
-        self.extensions.add_package(
-            Box::new(|mut response, request, _| {
-                let response: &mut Response<_> = unsafe { response.get_inner() };
-                let request: &FatRequest = unsafe { request.get_inner() };
+        struct Ext;
+        impl extensions::PackageCall for Ext {
+            fn call<'a>(
+                &'a self,
+                response: &'a mut Response<()>,
+                request: &'a FatRequest,
+                _: &'a Host,
+            ) -> RetFut<'a, ()> {
                 if request.uri().scheme_str() == Some("https") {
                     response
                         .headers_mut()
@@ -248,14 +326,25 @@ impl Host {
                 }
 
                 ready(())
-            }),
-            extensions::Id::new(8, "Adding HSTS header"),
-        );
+            }
+        }
+
+        self.extensions
+            .add_package(Box::new(Ext), extensions::Id::new(8, "Adding HSTS header"));
         self
     }
+
+    /// Add an alternative name to this host.
+    ///
+    /// See [the fiels](Self::alternative_names) for more details.
+    pub fn add_alternative_name(&mut self, name: impl AsRef<str>) -> &mut Self {
+        self.alternative_names.push(name.as_ref().to_owned());
+        self
+    }
+
     /// Disables client cache on this host.
     ///
-    /// This makes all [`ClientCachePreference`]s `no-store`.
+    /// This makes all [`comprash::ClientCachePreference`]s `no-store`.
     /// Use Kvarn extensions' `force_cache` to force certain files to cache.
     pub fn disable_client_cache(&mut self) -> &mut Self {
         self.options.disable_client_cache();
@@ -301,21 +390,46 @@ impl Host {
     pub(crate) fn is_secure(&self) -> bool {
         false
     }
+
+    /// Set the brotli compression level. 1-10, lower values are faster, but compress less.
+    ///
+    /// See [some benchmarks](https://quixdb.github.io/squash-benchmark/#results) for more context.
+    #[cfg(feature = "br")]
+    #[inline]
+    pub fn set_brotli_level(&mut self, level: u32) -> &mut Self {
+        self.compression_options.brotli_level = level;
+        self
+    }
+    /// Set the gzip compression level. 1-10, lower values are faster, but compress less.
+    ///
+    /// See [some benchmarks](https://quixdb.github.io/squash-benchmark/#results) for more context.
+    #[cfg(feature = "gzip")]
+    #[inline]
+    pub fn set_gzip_level(&mut self, level: u32) -> &mut Self {
+        self.compression_options.gzip_level = level;
+        self
+    }
 }
 impl Debug for Host {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
-        let mut d = f.debug_struct("Host");
-        d.field("name", &self.name);
-        #[cfg(feature = "https")]
-        d.field("certificate", &"[internal certificate]".as_clean());
-        d.field("path", &self.path);
-        d.field("extensions", &"[internal extension data]".as_clean());
-        d.field("file_cache", &"[internal cache]".as_clean());
-        d.field("response_cache", &"[internal cache]".as_clean());
-        d.field("limiter", &self.limiter);
-        d.field("vary", &self.vary);
-        d.field("options", &self.options);
-        d.finish()
+        let mut s = f.debug_struct(utils::ident_str!(Host));
+        utils::fmt_fields!(
+            s,
+            (self.name),
+            (self.alternative_names),
+            #[cfg(feature = "https")]
+            (self.certificate, &"[internal certificate]".as_clean()),
+            (self.path),
+            (self.extensions, &"[internal extension data]".as_clean()),
+            (self.file_cache, &"[internal cache]".as_clean()),
+            (self.response_cache, &"[internal cache]".as_clean()),
+            (self.limiter),
+            (self.vary),
+            (self.options),
+            (self.compression_options),
+        );
+
+        s.finish()
     }
 }
 /// Options for [`Host`].
@@ -325,6 +439,7 @@ impl Debug for Host {
 #[derive(Debug, Clone)]
 #[must_use]
 pub struct Options {
+    // # Miscellaneous
     /// Will be the default for folders; `/js/` will resolve to `/js/<folder_default>`.
     /// E.g. `/posts/` -> `/posts/index.html`
     ///
@@ -335,18 +450,27 @@ pub struct Options {
     ///
     /// If no value is passed, `html` is assumed.
     pub extension_default: Option<String>,
+    /// Default data directory for public files.
+    /// Default is `public`.
+    pub public_data_dir: Option<PathBuf>,
+    /// Default directory for overriding HTTP error responses.
+    /// Default is `errors`.
+    pub errors_dir: Option<PathBuf>,
+
+    // # Cache
     /// Returns `cache-control` header to be `no-store` by default, if enabled.
     ///
     /// Useful if you have a developing site and don't want traditionally static content to be in the client cache.
     pub disable_client_cache: bool,
-    /// Default data directory for public files.
-    /// Default is `public`
-    pub public_data_dir: Option<PathBuf>,
-
     /// Disables further caching by sending a [`StatusCode::NOT_MODIFIED`] when the
     /// `if-modified-since` header is sent and the resource is fresh.
     pub disable_if_modified_since: bool,
+    /// Filter to not cache certain [`StatusCode`]s.
+    ///
+    /// See [`CacheAction`] and [`default_status_code_cache_filter`] for more info.
+    pub status_code_cache_filter: fn(StatusCode) -> CacheAction,
 
+    // # Extensions
     /// Disables file system access for public files.
     ///
     /// This still enables custom error messages and reading of files through extensions.
@@ -356,19 +480,24 @@ impl Options {
     /// Creates a new [`Options`] with default settings.
     ///
     /// All [`Option`]s are [`None`] and all booleans are `false`.
+    /// [`Self::status_code_cache_filter`] uses [`default_status_code_cache_filter`].
     pub fn new() -> Self {
         Self {
             folder_default: None,
             extension_default: None,
-            disable_client_cache: false,
             public_data_dir: None,
+            errors_dir: None,
+
+            disable_client_cache: false,
             disable_if_modified_since: false,
+            status_code_cache_filter: default_status_code_cache_filter,
+
             disable_fs: false,
         }
     }
     /// Disables client cache on this host.
     ///
-    /// This makes all [`ClientCachePreference`]s `no-store`.
+    /// This makes all [`comprash::ClientCachePreference`]s `no-store`.
     /// Use Kvarn extensions' `force_cache` to force certain files to cache.
     pub fn disable_client_cache(&mut self) -> &mut Self {
         self.disable_client_cache = true;
@@ -381,12 +510,47 @@ impl Options {
         self.disable_fs = true;
         self
     }
-    /// Sets the relative directory (from the [`Host::path`]) to fetch data for the web in.
-    /// Defaults to `public`
+    /// Sets the directory (relative to the [`Host::path`]) to fetch data for the web in.
+    /// Defaults to `public`.
     pub fn set_public_data_dir(&mut self, path: impl AsRef<Path>) -> &mut Self {
         self.public_data_dir = Some(path.as_ref().to_path_buf());
         self
     }
+    /// Sets the directory (relative to the [`Host::path`]) to get HTTP error overrides from.
+    /// Defaults to `errors`.
+    pub fn set_errors_dir(&mut self, path: impl AsRef<Path>) -> &mut Self {
+        self.errors_dir = Some(path.as_ref().to_path_buf());
+        self
+    }
+
+    /// Gets the [`Self::folder_default`], as used by Kvarn.
+    /// Uses the default specified there.
+    #[must_use]
+    pub fn get_folder_default(&self) -> &str {
+        self.folder_default.as_deref().unwrap_or("index.html")
+    }
+    /// Gets the [`Self::extension_default`], as used by Kvarn.
+    /// Uses the default specified there.
+    #[must_use]
+    pub fn get_extension_default(&self) -> &str {
+        self.extension_default.as_deref().unwrap_or("html")
+    }
+    /// Gets the [`Self::public_data_dir`], as used by Kvarn.
+    /// Uses the default specified there.
+    #[must_use]
+    pub fn get_public_data_dir(&self) -> &Path {
+        self.public_data_dir
+            .as_deref()
+            .unwrap_or_else(|| Path::new("public"))
+    }
+    /// Gets the [`Self::errors_dir`], as used by Kvarn.
+    /// Uses the default specified there.
+    #[must_use]
+    pub fn get_errors_dir(&self) -> &Path {
+        self.public_data_dir
+            .as_deref()
+            .unwrap_or_else(|| Path::new("errors"))
+    }
 }
 impl Default for Options {
     fn default() -> Self {
@@ -394,20 +558,29 @@ impl Default for Options {
     }
 }
 
-/// A builder of [`Data`]. See [`Data::builder()`].
+/// A builder of [`Collection`]. See [`Collection::builder()`].
 #[derive(Debug)]
 #[must_use]
-pub struct DataBuilder(Data);
-impl DataBuilder {
+pub struct CollectionBuilder(Collection);
+impl CollectionBuilder {
     /// Adds `host` to the builder.
     /// This will match the `host` header and SNI hostname for [`Host.name`].
     #[inline]
     pub fn insert(mut self, host: Host) -> Self {
         self.check_secure(&host);
         if self.0.first.is_none() {
-            self.0.first = Some(host.name);
+            self.0.first = Some(host.name.clone());
         }
-        self.0.by_name.insert(host.name, host);
+        // it's important to insert the alt-names first, as they might contain the main name.
+        // If that happens, the inserted reference is overridden below.
+        for alt_name in &host.alternative_names {
+            self.0
+                .by_name
+                .insert(alt_name.clone(), HostValue::Ref(host.name.clone()));
+        }
+        self.0
+            .by_name
+            .insert(host.name.clone(), HostValue::Host(host));
         self
     }
     /// Adds a default `host` which is the fallback for all requests with a requested host
@@ -417,7 +590,7 @@ impl DataBuilder {
     /// > for requests to `localhost`.
     ///
     /// **NOTE:** This should be used with care as all secure connections to this server
-    /// with a SNI hostname that is not registered in this [`Data`], the client will get
+    /// with a SNI hostname that is not registered in this [`Collection`], the client will get
     /// a security warning, as the certificate of the default host is used.
     /// This is fine with HTTP, as that only delivers the website.
     ///
@@ -431,9 +604,8 @@ impl DataBuilder {
             self.0.default.is_none(),
             "Can not set default host multiple times."
         );
-        self.check_secure(&host);
-        self.0.default = Some(host);
-        self
+        self.0.default = Some(host.name.clone());
+        self.insert(host)
     }
     fn check_secure(&mut self, host: &Host) {
         if host.is_secure() {
@@ -447,44 +619,61 @@ impl DataBuilder {
         self.0.pre_host_limiter = limiter;
         self
     }
-    /// Puts the inner [`Data`] in a [`Arc`] and returns it.
+    /// Puts the inner [`Collection`] in a [`Arc`] and returns it.
     ///
     /// This works great with the overall flow of Kvarn. See [`RunConfig::execute()`] for an example.
     #[inline]
-    pub fn build(self) -> Arc<Data> {
+    #[must_use]
+    pub fn build(self) -> Arc<Collection> {
         Arc::new(self.into_inner())
     }
-    /// Converts `self` to a [`Data`].
+    /// Converts `self` to a [`Collection`].
     #[inline]
-    pub fn into_inner(self) -> Data {
+    pub fn into_inner(self) -> Collection {
         self.0
     }
 }
 
-/// A collection of [`Host`]s, with exactly one default and
-/// arbitrarily many other, indexed by [`Host.name`].
+#[derive(Debug)]
+#[allow(clippy::large_enum_variant)] // we want direct access to the host.
+enum HostValue {
+    Host(Host),
+    Ref(String),
+}
+impl HostValue {
+    fn as_host(&self) -> Option<&Host> {
+        match self {
+            Self::Host(h) => Some(h),
+            Self::Ref(_) => None,
+        }
+    }
+}
+
+/// A collection of [`Host`]s, with an optional default and
+/// arbitrarily many others, indexed by [`Host.name`].
 ///
-/// If only a default is specified, all requests, (e.g. those who lack a `host` header,
-/// have none, or all other values of the header) are channelled to the default.
+/// Tries to route to the host with it's name.
+/// If no host with a matching name is found, it'll fall back to [`default`](Self::get_default), if
+/// that's [`Some`].
 ///
-/// > When called as `localhost`, the first is always used.
+/// > When called from `localhost`, the first host added is always used.
 ///
-/// If the feature `https` is enabled, [`rustls::server::ResolvesServerCert`] in implemented
-/// using this default and host name pattern.
+/// If the feature `https` is enabled, [`rustls::server::ResolvesServerCert`] is implemented
+/// using the pattern described above.
 #[derive(Debug)]
 #[must_use]
-pub struct Data {
-    default: Option<Host>,
-    by_name: HashMap<&'static str, Host>,
-    first: Option<&'static str>,
+pub struct Collection {
+    default: Option<String>,
+    by_name: HashMap<String, HostValue>,
+    first: Option<String>,
     has_secure: bool,
     pre_host_limiter: LimitManager,
 }
-impl Data {
-    /// Creates a new [`DataBuilder`] with `default_host` as the default.
+impl Collection {
+    /// Creates a new [`CollectionBuilder`] with `default_host` as the default.
     #[inline]
-    pub fn builder() -> DataBuilder {
-        DataBuilder(Self {
+    pub fn builder() -> CollectionBuilder {
+        CollectionBuilder(Self {
             default: None,
             by_name: HashMap::new(),
             first: None,
@@ -496,57 +685,75 @@ impl Data {
     /// The host is the default. See [`host`] for more info.
     #[inline]
     pub fn simple_non_secure(default_host_name: &'static str, extensions: Extensions) -> Self {
-        Self {
-            default: Some(Host::unsecure(
+        Self::builder()
+            .default(Host::unsecure(
                 default_host_name,
                 ".",
                 extensions,
                 Options::default(),
-            )),
-            by_name: HashMap::new(),
-            first: None,
-            has_secure: false,
-            pre_host_limiter: LimitManager::default(),
-        }
+            ))
+            .into_inner()
     }
 
     /// Returns a reference to the default [`Host`].
     ///
-    /// Use [`Data::smart_get`] to get the appropriate host.
+    /// Use [`Self::get_from_request`] to get the appropriate host.
     #[inline]
+    #[must_use]
     pub fn get_default(&self) -> Option<&Host> {
-        self.default.as_ref()
+        self.default
+            .as_ref()
+            .and_then(|default| self.get_host(default))
     }
-    /// Gets a [`Host`] by name.
+    /// Get a [`Host`] by name.
     #[inline]
-    pub fn get_host(&self, host: &str) -> Option<&Host> {
-        self.by_name.get(host)
+    #[must_use]
+    pub fn get_host(&self, name: &str) -> Option<&Host> {
+        match self.by_name.get(name) {
+            Some(v) => match v {
+                HostValue::Host(h) => Some(h),
+                HostValue::Ref(r) => Some(
+                    self.by_name
+                        .get(r)
+                        .and_then(HostValue::as_host)
+                        .expect("internal error when resolving host"),
+                ),
+            },
+            None => None,
+        }
     }
-    /// Gets a [`Host`] by name, and returns the [`default`](Data::get_default) if none were found.
+    /// Get a [`Host`] by name, and returns the [`default`](Self::get_default) if none were found.
     #[inline]
-    pub fn get_or_default(&self, host: &str) -> Option<&Host> {
-        self.get_host(host)
+    #[must_use]
+    pub fn get_or_default(&self, name: &str) -> Option<&Host> {
+        self.get_host(name)
+            .or_else(|| name.strip_suffix('.').and_then(|name| self.get_host(name)))
             .or_else(|| self.get_default())
             .or_else(|| {
-                if host.split(':').next() == Some("localhost") {
-                    self.first.and_then(|host| self.get_host(host))
+                let base_host = name.split(':').next();
+                if base_host == Some("localhost")
+                    || base_host == Some("127.0.0.1")
+                    || base_host == Some("::1")
+                {
+                    self.first.as_ref().and_then(|host| self.get_host(host))
                 } else {
                     None
                 }
             })
     }
-    /// Gets a [`Host`] by name, if any, and returns it or the [`default`](Data::get_default)
-    /// if `maybe_host` is [`None`] or [`Data::get_or_default`] returns [`None`].
+    /// Get a [`Host`] by name, if any, and returns it or the [`default`](Self::get_default)
+    /// if `name` is [`None`] or [`Self::get_or_default`] returns [`None`].
     #[inline]
-    pub fn maybe_get_or_default(&self, maybe_host: Option<&str>) -> Option<&Host> {
-        match maybe_host {
+    #[must_use]
+    pub fn get_option_or_default(&self, name: Option<&str>) -> Option<&Host> {
+        match name {
             Some(host) => self.get_or_default(host),
             None => self.get_default(),
         }
     }
-    /// Cleverly gets the host depending on [`header::HOST`] and the `sni_hostname`.
+    /// Get the host depending on [`header::HOST`] and the `sni_hostname`.
     #[inline]
-    pub fn smart_get<'a>(
+    pub fn get_from_request<'a>(
         &'a self,
         request: &Request<Body>,
         sni_hostname: Option<&str>,
@@ -560,18 +767,19 @@ impl Data {
 
         let host = sni_hostname.or_else(|| get_header(request.headers()));
 
-        self.maybe_get_or_default(host)
+        self.get_option_or_default(host)
     }
 
     /// Returns if any [`Host`]s are [`Host::is_secure`].
     #[inline]
+    #[must_use]
     pub fn has_secure(&self) -> bool {
         self.has_secure
     }
 
-    /// Makes a [`rustls::ServerConfig`] from [`Data`].
+    /// Makes a [`rustls::ServerConfig`] from [`Self`].
     ///
-    /// This takes [`Data`] in an [`Arc`] and clones it.
+    /// This takes [`Self`] in an [`Arc`] and clones it.
     ///
     /// You should not have to call this, since [`PortDescriptor::new`] and [`PortDescriptor::https`] calls it internally.
     /// Though, you could use the [`host`] system by itself, without the rest of Kvarn.
@@ -596,22 +804,14 @@ impl Data {
     /// Clears all response caches.
     #[inline]
     pub async fn clear_response_caches(&self) {
-        // Handle default host
-        if let Some(cache) = self
-            .default
-            .as_ref()
-            .and_then(|h| h.response_cache.as_ref())
-        {
-            cache.lock().await.clear();
-        }
-        // All other
-        for host in self.by_name.values() {
+        for host in self.by_name.values().filter_map(HostValue::as_host) {
             if let Some(cache) = &host.response_cache {
-                cache.lock().await.clear();
+                cache.write().await.clear();
             }
         }
     }
     /// Clears a single `uri` in `host`.
+    /// If `host` is `""` or `"default"`, the [default](Self::get_default) host is used.
     ///
     /// # Returns
     ///
@@ -619,20 +819,20 @@ impl Data {
     ///
     /// This will probably become a error enum in the future.
     ///
-    /// It will lever return (false, true).
+    /// It will never return (false, true).
     pub async fn clear_page(&self, host: &str, uri: &Uri) -> (bool, bool) {
         let key = UriKey::path_and_query(uri);
 
         let mut found = false;
         let mut cleared = false;
         if host.is_empty() || host == "default" {
-            found = true;
             if let Some(cache) = self
-                .default
+                .get_default()
                 .as_ref()
                 .and_then(|h| h.response_cache.as_ref())
             {
-                let mut lock = cache.lock().await;
+                found = true;
+                let mut lock = cache.write().await;
                 if key
                     .call_all(|key| lock.remove(key).into_option())
                     .1
@@ -641,10 +841,10 @@ impl Data {
                     cleared = true;
                 }
             }
-        } else if let Some(host) = self.by_name.get(host) {
+        } else if let Some(host) = self.get_host(host) {
             found = true;
             if let Some(cache) = &host.response_cache {
-                let mut lock = cache.lock().await;
+                let mut lock = cache.write().await;
                 if key
                     .call_all(|key| lock.remove(key).into_option())
                     .1
@@ -659,56 +859,61 @@ impl Data {
     /// Clears all file caches.
     #[inline]
     pub async fn clear_file_caches(&self) {
-        if let Some(cache) = self.default.as_ref().and_then(|h| h.file_cache.as_ref()) {
-            cache.lock().await.clear();
-        }
-        for host in self.by_name.values() {
+        for host in self.by_name.values().filter_map(HostValue::as_host) {
             if let Some(cache) = &host.file_cache {
-                cache.lock().await.clear();
+                cache.write().await.clear();
             }
         }
     }
-    /// Clears the `path` from all caches.
+    /// Clears a single `path` in `host`.
+    /// If `host` is `""` or `"default"`, the [default](Self::get_default) host is used.
     ///
-    /// This iterates over all caches and [locks](Mutex::lock) them, which takes a lot of time.
+    /// # Returns
+    ///
+    /// (if host was found, cleared page)
+    ///
+    /// This will probably become a error enum in the future.
+    ///
+    /// It will never return (false, true).
+    ///
+    /// This iterates over all caches and [locks](RwLock::write) them, which takes a lot of time.
     /// Though, it's not blocking.
-    pub async fn clear_file_in_cache<P: AsRef<Path>>(&self, path: &P) -> bool {
+    pub async fn clear_file(&self, host: &str, path: impl AsRef<Path>) -> (bool, bool) {
+        let path = path.as_ref();
         let mut found = false;
-        if let Some(cache) = self.default.as_ref().and_then(|h| h.file_cache.as_ref()) {
-            if cache
-                .lock()
-                .await
-                .remove(path.as_ref())
-                .into_option()
-                .is_some()
+        let mut cleared = false;
+        if host.is_empty() || host == "default" {
+            if let Some(cache) = self
+                .get_default()
+                .as_ref()
+                .and_then(|h| h.file_cache.as_ref())
             {
                 found = true;
+                let mut lock = cache.write().await;
+                if lock.remove(path).into_option().is_some() {
+                    cleared = true;
+                }
             }
-        }
-        for host in self.by_name.values() {
+        } else if let Some(host) = self.get_host(host) {
+            found = true;
             if let Some(cache) = &host.file_cache {
-                if cache
-                    .lock()
-                    .await
-                    .remove(path.as_ref())
-                    .into_option()
-                    .is_some()
-                {
-                    found = true;
+                let mut lock = cache.write().await;
+                if lock.remove(path).into_option().is_some() {
+                    cleared = true;
                 }
             }
         }
-        found
+        (found, cleared)
     }
 }
 #[cfg(feature = "https")]
-impl ResolvesServerCert for Data {
+impl ResolvesServerCert for Collection {
     #[inline]
     fn resolve(&self, client_hello: ClientHello<'_>) -> Option<Arc<sign::CertifiedKey>> {
         // Mostly returns true, since we have a default
         // Will however return false if certificate is not present
         // in found host or default host.
-        self.maybe_get_or_default(client_hello.server_name())
+        self.get_option_or_default(client_hello.server_name())
             .and_then(|host| host.certificate.as_ref())
             .cloned()
     }
@@ -717,7 +922,7 @@ impl ResolvesServerCert for Data {
 /// All the supported ALPN protocols.
 ///
 /// > ***Note:** this is often not needed, as the ALPN protocols
-/// are set in [`host::Data::make_config()`].*
+/// are set in [`host::Collection::make_config()`].*
 #[must_use]
 pub fn alpn() -> Vec<Vec<u8>> {
     #[allow(unused_mut)]
@@ -729,6 +934,50 @@ pub fn alpn() -> Vec<Vec<u8>> {
     vec
 }
 
+/// Per host filter output of whether or not to cache a response with some [`StatusCode`].
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+#[must_use]
+pub enum CacheAction {
+    /// Cache this status code response.
+    Cache,
+    /// Don't cache, in hope following responses have status codes which can be cached.
+    Drop,
+}
+impl CacheAction {
+    /// Returns [`Self::Cache`] if `cache` is true. Else [`Self::Drop`].
+    pub fn from_cache(cache: bool) -> Self {
+        if cache {
+            Self::Cache
+        } else {
+            Self::Drop
+        }
+    }
+    /// Returns [`Self::Drop`] if `drop` is true. Else [`Self::Cache`].
+    pub fn from_drop(drop: bool) -> Self {
+        Self::from_cache(!drop)
+    }
+    /// Returns true if `self` is [`Self::Cache`].
+    #[must_use]
+    pub fn into_cache(self) -> bool {
+        matches!(self, Self::Cache)
+    }
+    /// Returns true if `self` is [`Self::Drop`].
+    #[must_use]
+    pub fn into_drop(self) -> bool {
+        matches!(self, Self::Drop)
+    }
+}
+
+/// This is the default for [`Options::status_code_cache_filter`].
+///
+/// This caches the request on every [`StatusCode`] except
+/// - client errors, however `404 Not Found` and `410 Gone` are still cached
+/// - any [informational response](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status#information_responses)
+/// - [304 Not Modified](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/304)
+pub fn default_status_code_cache_filter(code: StatusCode) -> CacheAction {
+    CacheAction::from_drop(matches!(code.as_u16(), 400..=403 | 405..=409 | 411..=499|100..=199|304))
+}
+
 /// An error regarding creation of a [`rustls::sign::CertifiedKey`].
 #[cfg(feature = "https")]
 #[derive(Debug)]
